=begin 
(c) TIG 2009
Script: triangulate.rb
Code Usage:  triangulatefaces(array_of_faces)
Result: it triangulates faces in array...
Plugin Usage: 
Result: it triangulates faces in selection...
Known Issues:
  Even complex single looped faces should work OK.
  Simple faces with one or two holes are also usually OK.
  Very large numbers of faces might cause a 'white-out', 
    processing is continuing, please be patient...
  Complex faces with complex multiple holes in them might 'splat' !
    Add some extra dividing lines and retry...
  Very small faces might not work properly and 'splat' ! Scale them...
  Failed triangulations are warned and can be undone...
  Save your work before using this tool...

v1.0 First issue. 20090703
v1.1 4 sided faces with 3 in line vertices trapped.
     undo operation added.  20090703 pm
v1.2 L-shaped AND 4-sided faces trapped. 20090703 pm+
v1.3 Menu added, warns if no suitable selection. 20090703 pm++
v2.0 Now triangulates all >= 4-sided faces and traps awkward L/U/holed 
     shapes etc. 20090707
v2.1 Renamed 'triangulatefaces' to avoid clashes with delauney/cloud etc
=end
### Code ###
def triangulatefaces(array_of_faces)
 ### checker to see if triangulating newline falls wholly within face
 def line_on_face?(point0,point1,face)
  vector=point1-point0
  distance=point0.distance(point1)
  onface=true; offset=1.0.mm
  test_point=point0.offset(vector,offset);onface=false if face.classify_point(test_point) != 1 ### fails near start
  test_point=point1.offset(vector.reverse,offset);onface=false if face.classify_point(test_point) != 1 ### fails near end
  test_point=point0.offset(vector,distance/2);onface=false if face.classify_point(test_point) != 1 ### it fails in middle
  offset=2.0.mm
  while onface and offset<(distance-(offset*2))
    test_point=point0.offset(vector,offset)
    onface=false if face.classify_point(test_point) != 1
    offset=offset+1.0.mm
  end#while ### it can fail along line elsewhere !
  #onface=true if distance <=0.2.mm
  return onface
 end#def line_on_face?
 
 ### main code...
  model=Sketchup.active_model
  ents=model.active_entities
  model.start_operation("Triangulate Faces")
  faces=array_of_faces
  faces4=[]
  faces.each{|e|faces4.push(e)if e.kind_of?(Sketchup::Face) and e.edges.length>=4}
  tick=0;tickX=0;trytick=0
  while faces4[0]
    msg="Triangulating"
    faces4.each{|face|
      
      Sketchup::set_status_text(msg,SB_PROMPT)
      if tick==tickX
        msg=msg+". "
        tickX+=16
      end#if
      tick+=1
      
      if face.loops.length==1 ### only has an outer loop
        #UI.messagebox"Outer Loop"
        vertices=face.outer_loop.vertices
        points=[]
        vertices.each{|v|points.push(v.position)}
        points=points.push(points[0].clone,points[1].clone)     
        newline=nil;count=0
        while not newline and count<=points.length-3 #and points.length>5
          if line_on_face?(points[count],points[count+2],face) 
            newline=ents.add_line(points[count],points[count+2])
            newline.faces.each{|f|
              faces4=faces4-[f]if f.edges.length==3
              faces4.push(f)if f.edges.length>3
              faces4.uniq!
            }
          end#if
          count+=1
        end#while

      else ### must deal will inner loops too...
        #UI.messagebox"Inner Loop"
        ### face.classify_point(pt) can fail with inner loops
        ### as it can still think the point is on face when it's not !!!
        points=[]
        face.outer_loop.vertices.each{|v|points.push(v.position)}
        loops=face.loops-[face.outer_loop]
        loops.each{|loop|lpoints=[]
          loop.vertices.each{|v|lpoints.push(v.position)}
          newlines=[]
          lpoints.each{|lpoint|
            dis=lpoint.distance(points[0])
            pointX=points[0]
            points.each{|point|
              if not point==lpoint and point.distance(lpoint)<dis
                pointX=point
              end#if
            }
            if line_on_face?(pointX,lpoint,face) and not newlines[0]
              newline=ents.add_line(pointX,lpoint)
              newline.faces.each{|f|
                faces4=faces4-[f]if f.edges.length==3
                faces4.push(f)if f.edges.length>3
                faces4.uniq!
              }
              newlines.push(newline)
              ### so only one newline added so loop added to outer_loop
            end#if
          }
        }
        
      end#if
      
    }
    trytick+=1
    if trytick>100
      faces4=[]
      if UI.messagebox("Probable Triangulating Error !\nUndo ?",MB_YESNO,"")==6 ### 6=YES 7=NO
        Sketchup.send_action("editUndo:")
      end#if          
    end#if

  end#while
  Sketchup::set_status_text("",SB_PROMPT)
  model.commit_operation
  return nil
end#def
### Menu ###
if not file_loaded?(File.basename(__FILE__))
  UI.menu("Plugins").add_item("Triangulate Faces in Selection"){triangulatefaces(Sketchup.active_model.selection.to_a);Sketchup.active_model.selection.clear}
end#if
file_loaded(File.basename(__FILE__))